# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import pandas as pd, numpy as np
import dataiku
from dataiku.core.sql import SQLExecutor2
from string import Template

# NOTE: You can check the job log to see the results of any print() calls


##################################################################################################
# Reading Data
##################################################################################################


###########################################
# Load Data from Dataset into a DataFrame
###########################################

input_ds = dataiku.Dataset("INPUT")
input_df = input_ds.get_dataframe(infer_with_pandas=False)

# NOTES:
#
# > Recommend infer_with_pandas=False so data types for Pandas DataFrame columns are set per the
#   DSS Schema (which in turn may be set from SQL table). This assures that columns that are mostly
#   but not always numeric or have leading zeros are consistently loaded correctly.
#
# > Avoid using columns argument as it currently doesn’t work properly.
#

print('***Load Data from Dataset into a DataFrame***')
print('input_df data types:')
print(input_df.dtypes)


###########################################
# Load SQL Query Output into DataFrame
###########################################

# Get project variables
variables = dataiku.get_custom_variables() # sqlFilterValue variable defined through UI

# Get SQL table name and add to variables for use below
# Note that any variables in the table name (such as ${projectKey}) have been resolved in
# the value returned by get_location_info().
variables['sqlTableName'] = input_ds.get_location_info()['info']['table'] 

# Specify SQL query referencing any needed project variables
query_with_vars = """
SELECT
    INTEGER_COL,
    CAST(INTEGER_COL * 2 AS BIGINT) AS ANOTHER_BIGINT_COL,
    'SQL Query' AS SOURCE 
FROM ${sqlTableName} 
WHERE INTEGER_COL <= ${sqlFilterValue}
"""

# Replace variables in query (Template class works with DSS variable references)
query = Template(query_with_vars).substitute(variables)

# Execute the query
executor = SQLExecutor2(dataset='INPUT') 
sql_query_df = executor.query_to_df(query, infer_from_schema=True)

# NOTES:
#
# > Infer_from_schema = True does the same thing as get_dataframe’s infer_with_pandas = False
#   i.e., set the DataFrame data types per the DSS Schema.
# 
# > The above approach is designed to enable use of project variables and to be resilient
#   to changes in the project key (which would occur if project is copied) and to a change
#   in the table name. In some cases you may not need this additional functionality. Here's
#   the minimal option (as one liner):
#
#   df = SQLExecutor2(dataset='INPUT').query_to_df("SELECT * FROM PYSQL_INPUT", infer_from_schema=True)
#

print('***Load SQL Query Output into DataFrame***')
print('SQL Query returned {1} columns and {0} rows.'.format(*sql_query_df.shape))


###########################################
# Memory Optimized Data Loading
###########################################

# Get dataset schema 
schema = input_ds.read_schema()

# Set names, dtypes, and dtcols variables from schema (these are used later by forced types load)
(names, dtypes, dtcols) = input_ds.get_dataframe_schema_st(schema, infer_with_pandas=False)

# Update dtypes to use memory efficient data types
# Note that only string columns with many common values should be converted to the category data type;
# if string columns have unique values or mostly unique values (say more than 50% of total rows), then
# those columns should be left as string rather than converted to category.
type_map = (('tinyint', np.int8), ('smallint', np.int16), ('string', 'category'))
for colspec in schema:
    for dss_type, pandas_type in type_map:
        if colspec['type'] == dss_type:
            dtypes[colspec['name']] = pandas_type

# Load dataframe using updated dtypes
# iter_dataframes_forced_types enables setting of dtypes so using this with 1 TB
# chunk size to load the dataframe in one step
mem_efficient_df = next(input_ds.iter_dataframes_forced_types(names, dtypes, dtcols, chunksize=1E12))


print('***Memory Optimized Data Loading***')
print('mem_efficient_df data types:')
print(mem_efficient_df.dtypes)


##################################################################################################
# Writing Data
##################################################################################################


###########################################
# Populating a Dataset from a DataFrame
###########################################

output_no_schema_ds = dataiku.Dataset('OUTPUT_NO_SCHEMA')
output_no_schema_ds.write_with_schema(mem_efficient_df)

# NOTES
#
# > Data types for the DSS schema and the SQL table data types are inferred from the Pandas dataframe.
#
#   This is easiest option and will often work just fine. 
#
# > Can of course set dataframe column types as desired so can avoid some of the previously mentioned 
#   load data issues; however:
#   - Tinyint (int8) and smallint (int16) columns currently set to VARCHAR(500)
#   - String (object) columns set to VARCHAR(500) 
#
#   This data type assignments may work just fine. We sometimes write tables with a large number of
#   columns and too many VARCHAR(500) columns would cause the record size to exceed the database 
#   limit.
#
# > The SQL table is dropped and recreated each time.
# 

print('***Populating a Dataset from a DataFrame***')
print('OUTPUT_NO_SCHEMA DSS Schema data types:')
print(output_no_schema_ds.read_schema())

###########################################
# Populating a Dataset with a Defined Schema
###########################################

output_schema_ds = dataiku.Dataset('OUTPUT_SCHEMA')
output_schema_ds.write_dataframe(input_df)

print('***Populating a Dataset with a Defined Schema***')
print('OUTPUT_SCHEMA DSS Schema data types:')
print(output_schema_ds.read_schema())

# NOTES
#
# > Use write_dataframe for full control over output SQL table data types
# 
# > Preset dataset schema (and thereby SQL table data types)
#   - In UI under dataset Settings, Schema tab
#   - Programmatically via set_schema() (use get_schema() to see structure)
#
# > Dataframe  types must be compatible with the SQL table data types


###########################################
# Appending Rows to an Existing Table
###########################################

output_schema_append_ds = dataiku.Dataset('OUTPUT_SCHEMA_APPEND')
output_schema_append_ds.write_dataframe(input_df)

# NOTES
#
# > As you can see, this is simply a variation on “Populating a Dataset with
#   a Defined Schema” (the Python code is the same).
# 
# > To set up appending, check "Append instead of overwrite“ box for the 
#   appropriate dataset on recipe Inputs/Outputs tab.
# 
# > Appending not allowed in a Jupyter Notebook
#
# > Note that for some use cases, appending records directly may not the best option.
#
#   Instead we often Write data to a SQL table that is dropped and recreated each
#   time and use a subsequent SQL recipe to append records to another table. The 
#   main advantage is reliability as SQL can easily handle duplicate records efficiently
#   that may occur when a process is run more than once by mistake or fails before
#   completing.

##################################################################################################
# Executing SQL
##################################################################################################


###########################################
# Populate a Dataset from a SQL Query
###########################################

# Must specify an input dataset as the SQL statements will be run under the dataset's connection
# Note there is no requirement to use the input dataset.

# Get reference to output dataset object
out_ds = dataiku.Dataset("OUTPUT_RECIPE_FRAGMENT")

# Will in pre_queries create a temp table of input records using a value from input table as a filter
pre_stmts = """
CREATE VOLATILE TABLE TMP_INPUT AS {0} WITH NO DATA ON COMMIT PRESERVE ROWS;
INSERT INTO TMP_INPUT SELECT * FROM {0} WHERE INTEGER_COL = {1}
"""
pre_stmts = pre_stmts.format(input_ds.get_location_info()['info']['table'], str(input_df.at[2, 'INTEGER_COL']))

# One list item per statement so avoid Teradata error "Only an ET or null statement is legal after a DDL Statement"
# With other databases (e.g., Netezza) can have multiple statements in one item list
pre_stmts = pre_stmts.split(';')


# Since create table statement executed twice (the first time to determine schema)
# need to drop the table here before pre_stmts executed the second time. A better
# approach would be to include DROP TABLE TMP_INPUT IF EXISTS as the first line in
# pre_stmts but unfortunately Teradata doesn't support this syntax.
post_stmts = ["DROP TABLE TMP_INPUT"] # must be a list

final_query = "SELECT * FROM TMP_INPUT" # must be a select statement

# Run the partial recipe (note that pre_queries and post_queries expect lists)
SQLExecutor2.exec_recipe_fragment(out_ds, final_query, pre_queries=pre_stmts, post_queries=post_stmts)


###########################################
# Execute Arbitrary SQL Statements
###########################################

# Clear tables, create tables, rename tables, update indexes, etc.

# Create a SQL Executor instance that will run under connection of the OUTPUT_SQL_STATEMENT dataset
executor = SQLExecutor2(dataset='OUTPUT_SQL_STATEMENT') # or connection='CONNECTION_NAME’

# Get reference to output dataset object (optional but will use here to get output SQL table name )
output_sql_stmt_ds = dataiku.Dataset("OUTPUT_SQL_STATEMENT")

# Set SQL statements to execute
# Some databases (e.g. Netezza) require a COMMIT as the last statement
pre_stmts = """
--DROP TABLE {0};
CREATE TABLE {0} AS {1} WITH NO DATA;
ALTER TABLE {0} MODIFY PRIMARY INDEX(BYTEINT_COL);
INSERT INTO {0} SELECT * FROM {1};
DELETE FROM {0} WHERE BYTEINT_COL = 1
"""

pre_stmts = pre_stmts.format(output_sql_stmt_ds.get_location_info()['info']['table'],
                             input_ds.get_location_info()['info']['table'])

# One list item per statement so avoid Teradata error "Only an ET or null statement is legal after a DDL Statement"
# With other databases (e.g., Netezza) can have multiple statements in one item list
pre_stmts = pre_stmts.split(';')

# Set final query (must have one) - can use to check status of statements. In this case pulling record
# count. Also fine to specify a dummy query like SELECT 1.
query =  "SELECT COUNT(*) AS REC_CNT FROM " + output_sql_stmt_ds.get_location_info()['info']['table']

# Run sql statements returning record count in output table
result_df = executor.query_to_df(query, pre_queries=pre_stmts)

# Check result (e.g., in a real situation might use an assert to confirm result is as expected)
print(result_df.at[0, 'REC_CNT'])
